#include "message_cli.h"

#include <zephyr/logging/log.h>
LOG_MODULE_DECLARE(message);

BUILD_ASSERT(BT_MESH_MODEL_BUF_LEN(MESSAGE_MODEL_OP_MESSAGE, MESSAGE_MODEL_MSG_MAXLEN_MESSAGE) <= BT_MESH_RX_SDU_MAX, "The message must fit inside an application SDU.");
BUILD_ASSERT(BT_MESH_MODEL_BUF_LEN(MESSAGE_MODEL_OP_MESSAGE, MESSAGE_MODEL_MSG_MAXLEN_MESSAGE) <= BT_MESH_TX_SDU_MAX, "The message must fit inside an application SDU.");
BUILD_ASSERT(BT_MESH_MODEL_BUF_LEN(MESSAGE_MODEL_OP_MESSAGE, MESSAGE_MODEL_BUFF_MAXLEN) <= BT_MESH_RX_SDU_MAX, "The buffer must fit inside an application SDU.");
BUILD_ASSERT(BT_MESH_MODEL_BUF_LEN(MESSAGE_MODEL_OP_MESSAGE, MESSAGE_MODEL_BUFF_MAXLEN) <= BT_MESH_TX_SDU_MAX, "The buffer must fit inside an application SDU.");


static void encode_status(struct net_buf_simple *buf, enum message_model_status status)
{
	bt_mesh_model_msg_init(buf, MESSAGE_MODEL_OP_STATUS);
	net_buf_simple_add_u8(buf, status);
}

static const uint8_t *extract_msg(struct net_buf_simple *buf)
{
	buf->data[buf->len - 1] = '\0';
	return net_buf_simple_pull_mem(buf, buf->len);
}

static const uint8_t *extract_buffer(struct net_buf_simple *buf)
{
	return net_buf_simple_pull_mem(buf, buf->len);
}

static int handle_message(struct bt_mesh_model *model, struct bt_mesh_msg_ctx *ctx, struct net_buf_simple *buf)
{
	struct message_model_cli *cli = model->user_data;
	const uint8_t *msg;

	msg = extract_msg(buf);

	if (cli->handlers->message) {
		cli->handlers->message(cli, ctx, msg);
	}

	return 0;
}

static int handle_buffer(struct bt_mesh_model *model, struct bt_mesh_msg_ctx *ctx, struct net_buf_simple *buf)
{
	struct message_model_cli *cli = model->user_data;
	const uint8_t *buffer;
	uint16_t len = buf->len;
	buffer = extract_buffer(buf);

	/* Call the buffer callback */
	if (cli->handlers->buffer) {
		cli->handlers->buffer(cli, ctx, buffer, len);
	}

	return 0;
}

static void send_message_reply(struct message_model_cli *cli, struct bt_mesh_msg_ctx *ctx)
{
	BT_MESH_MODEL_BUF_DEFINE(msg, MESSAGE_MODEL_OP_MESSAGE_REPLY, MESSAGE_MODEL_MSG_LEN_MESSAGE_REPLY);
	bt_mesh_model_msg_init(&msg, MESSAGE_MODEL_OP_MESSAGE_REPLY);

	(void)bt_mesh_model_send(cli->model, ctx, &msg, NULL, NULL);
}

static int handle_private_message(struct bt_mesh_model *model, struct bt_mesh_msg_ctx *ctx, struct net_buf_simple *buf)
{
	struct message_model_cli *cli = model->user_data;
	const uint8_t *msg;

	msg = extract_msg(buf);

	if (cli->handlers->private_message) {
		cli->handlers->private_message(cli, ctx, msg);
	}

	send_message_reply(cli, ctx);
	return 0;
}

static int handle_message_reply(struct bt_mesh_model *model, struct bt_mesh_msg_ctx *ctx, struct net_buf_simple *buf)
{
	struct message_model_cli *cli = model->user_data;

	if (cli->handlers->message_reply) {
		cli->handlers->message_reply(cli, ctx);
	}

	return 0;
}

static int handle_status(struct bt_mesh_model *model, struct bt_mesh_msg_ctx *ctx, struct net_buf_simple *buf)
{
	struct message_model_cli *cli = model->user_data;
	enum message_model_status status;

	status = net_buf_simple_pull_u8(buf);

	if (cli->handlers->status) {
		cli->handlers->status(cli, ctx, status);
	}

	return 0;
}

static int handle_status_get(struct bt_mesh_model *model, struct bt_mesh_msg_ctx *ctx, struct net_buf_simple *buf)
{
	struct message_model_cli *cli = model->user_data;

	BT_MESH_MODEL_BUF_DEFINE(msg, MESSAGE_MODEL_OP_STATUS, MESSAGE_MODEL_MSG_LEN_STATUS);

	encode_status(&msg, cli->status);

	(void) bt_mesh_model_send(cli->model, ctx, &msg, NULL, NULL);

	return 0;
}

/* .. include_startingpoint_chat_cli_rst_2 */
const struct bt_mesh_model_op _message_model_cli_op[] = {
    {
		MESSAGE_MODEL_OP_MESSAGE,
		BT_MESH_LEN_MIN(MESSAGE_MODEL_MSG_MINLEN_MESSAGE),
		handle_message
	},
	{
		MESSAGE_MODEL_OP_PRIVATE_MESSAGE,
		BT_MESH_LEN_MIN(MESSAGE_MODEL_MSG_MINLEN_MESSAGE),
		handle_private_message
	},
	{
        MESSAGE_MODEL_OP_MESSAGE_REPLY,
		BT_MESH_LEN_EXACT(MESSAGE_MODEL_MSG_LEN_MESSAGE_REPLY),
		handle_message_reply
	},
	{
		MESSAGE_MODEL_OP_STATUS,
		BT_MESH_LEN_EXACT(MESSAGE_MODEL_MSG_LEN_STATUS),
		handle_status
	},
	{
		MESSAGE_MODEL_OP_STATUS_GET,
		BT_MESH_LEN_EXACT(MESSAGE_MODEL_MSG_LEN_STATUS_GET),
		handle_status_get
	},
	{
		MESSAGE_MODEL_OP_BUFFER,
		BT_MESH_LEN_MIN(MESSAGE_MODEL_BUFF_MINLEN),
		handle_buffer
	},
	BT_MESH_MODEL_OP_END,
};


static int message_model_update_handler(struct bt_mesh_model *model)
{
	struct message_model_cli *cli = model->user_data;

	/* Continue publishing current status. */
	encode_status(model->pub->msg, cli->status);

	return 0;
}


#ifdef CONFIG_BT_SETTINGS
static int message_model_settings_set(struct bt_mesh_model *model, const char *name, size_t len_rd, settings_read_cb read_cb, void *cb_arg)
{
	struct message_model_cli *cli = model->user_data;

	if (name) {
		return -ENOENT;
	}

	ssize_t bytes = read_cb(cb_arg, &cli->status, sizeof(cli->status));
	if (bytes < 0) {
		return bytes;
	}

	if (bytes != 0 && bytes != sizeof(cli->status)) {
		return -EINVAL;
	}

	return 0;
}
#endif

static int message_model_init(struct bt_mesh_model *model)
{
	struct message_model_cli *cli = model->user_data;

	cli->model = model;

	net_buf_simple_init_with_data(&cli->pub_msg, cli->buf, sizeof(cli->buf));
	cli->pub.msg = &cli->pub_msg;
	cli->pub.update = message_model_update_handler;

	return 0;
}

static int message_model_start(struct bt_mesh_model *model)
{
	struct message_model_cli *cli = model->user_data;

	if (cli->handlers->start) {
		cli->handlers->start(cli);
	}

	return 0;
}

static void message_model_reset(struct bt_mesh_model *model)
{
	struct message_model_cli *cli = model->user_data;

	cli->status = MESSAGE_MODEL_STATUS_DONE;

	if (IS_ENABLED(CONFIG_BT_SETTINGS)) {
		(void) bt_mesh_model_data_store(model, true, NULL, NULL, 0);
	}
}

const struct bt_mesh_model_cb _message_model_cli_cb = {
	.init = message_model_init,
	.start = message_model_start,
#ifdef CONFIG_BT_SETTINGS
	.settings_set = message_model_settings_set,
#endif
	.reset = message_model_reset,
};

int message_model_status_set(struct message_model_cli *cli, enum message_model_status status)
{
	if (status != cli->status) {
		cli->status = status;  
	}

	encode_status(cli->model->pub->msg, cli->status);

	return bt_mesh_model_publish(cli->model);
}

int message_model_status_get(struct message_model_cli *cli, uint16_t addr)
{
	struct bt_mesh_msg_ctx ctx = {
		.addr = addr,
		.app_idx = cli->model->keys[0],
		.send_ttl = BT_MESH_TTL_DEFAULT,
		.send_rel = true,
	};

	BT_MESH_MODEL_BUF_DEFINE(buf, MESSAGE_MODEL_OP_STATUS_GET, MESSAGE_MODEL_MSG_LEN_STATUS_GET);
	bt_mesh_model_msg_init(&buf, MESSAGE_MODEL_OP_STATUS_GET);

	return bt_mesh_model_send(cli->model, &ctx, &buf, NULL, NULL);
}

int message_model_message_send(struct message_model_cli *cli, const uint8_t *msg)
{
	struct net_buf_simple *buf = cli->model->pub->msg;

	bt_mesh_model_msg_init(buf, MESSAGE_MODEL_OP_MESSAGE);

	net_buf_simple_add_mem(buf, msg, strnlen(msg, MESSAGE_MODEL_MSG_MESSAGE_LENGTH));
	net_buf_simple_add_u8(buf, '\0');

	return bt_mesh_model_publish(cli->model);
}

int message_model_buffer_send(struct message_model_cli *cli, const uint8_t *buffer, int buff_len)
{
	struct net_buf_simple *buf = cli->model->pub->msg;

	bt_mesh_model_msg_init(buf, MESSAGE_MODEL_OP_BUFFER);

	net_buf_simple_add_mem(buf, buffer, buff_len);

	return bt_mesh_model_publish(cli->model);
}

int message_model_private_message_send(struct message_model_cli *cli, uint16_t addr, const uint8_t *msg)
{
	struct bt_mesh_msg_ctx ctx = {
		.addr = addr,
		.app_idx = cli->model->keys[0],
		.send_ttl = BT_MESH_TTL_DEFAULT,
		.send_rel = true,
	};

	BT_MESH_MODEL_BUF_DEFINE(buf, MESSAGE_MODEL_OP_PRIVATE_MESSAGE, MESSAGE_MODEL_MSG_MAXLEN_MESSAGE);
	bt_mesh_model_msg_init(&buf, MESSAGE_MODEL_OP_PRIVATE_MESSAGE);

	net_buf_simple_add_mem(&buf, msg, strnlen(msg, MESSAGE_MODEL_MSG_MESSAGE_LENGTH));
	net_buf_simple_add_u8(&buf, '\0');

	return bt_mesh_model_send(cli->model, &ctx, &buf, NULL, NULL);
}